@@ -1,9 +1,9 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2019-2023 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2019-2024 -- leonerd@leonerd.org.uk
 
-package Object::Pad 0.808;
+package Object::Pad 0.810;
 
 use v5.14;
 use warnings;
@@ -115,7 +115,7 @@ module to only silence the module's warnings selectively:
 
    use Object::Pad ':experimental(inherit_field)';
 
-   use Object::Pad ':experimental';  # all of the above
+   use Object::Pad ':experimental(:all)';  # all of the above
 
 I<Since version 0.64.>
 
@@ -124,6 +124,10 @@ in the parens, separated by spaces:
 
    use Object::Pad ':experimental(mop custom_field_attr)';
 
+I<Since version 0.810> attempting to request all of the experiments at once
+by using an empty C<:experimental()> is currently accepted, but yields a
+warning. This may be removed in future.
+
 =head2 Automatic Construction
 
 Classes are automatically provided with a constructor method, called C<new>,
@@ -345,6 +349,35 @@ code. It is recommended to enable this whenever possible, as a later version
 of this module will likely perform this behaviour unconditionally whenever no
 C<BUILD> blocks are present.
 
+=head2 class (anon)
+
+   my $class = class :ATTRS... { ... };
+
+I<Since version 0.809.>
+
+If a C<class> keyword is not followed by a package name, it creates an
+anonymous class expression. This is an expression that yields a value suitable
+to use as a constructor invocant for creating instances of that class, without
+specifying what its package name will actually be.
+
+This is useful for creating small one-off instances inline in expressions,
+such as in unit tests. Since it still accepts the usual attributes and inner
+body statements, it can be useful for creating one-off instances of roles,
+with required methods being applied.
+
+   my $testobj = (class {
+      apply Role::Under::Test;
+      method required { return "a useful value"; }
+   })->new;
+
+Due to limitations on how classes work in Perl, anonymous classes are still
+backed by long-lived named classes in the global symbol table, unlike true
+anonymous functions which can go out of scope and be reclaimed once no
+references to them remain in existence. This means that anonymous classes will
+retain references to any variables captured within them, even if the class
+expression itself goes out of scope and any instances created by it no longer
+remain.
+
 =head2 role
 
    role Name :ATTRS... {
@@ -1323,11 +1356,18 @@ sub _import_experimental
       my $sym = $syms->[$i];
 
       if( $sym eq ":experimental" ) {
+         carp "Enabling all Object::Pad experiments with an unqualified :experimental";
          $enabled{$_}++ for @experiments;
       }
       elsif( $sym =~ m/^:experimental\((.*)\)$/ ) {
-         my $tags = $1 =~ s/^\s+|\s+$//gr; # trim
-         $enabled{$_}++ for split m/\s+/, $tags;
+         foreach my $tag ( split m/\s+/, $1 =~ s/^\s+|\s+$//gr ) {
+            if( $tag eq ":all" ) {
+               $enabled{$_}++ for @experiments;
+            }
+            else {
+               $enabled{$tag}++;
+            }
+         }
       }
       else {
          $i++;
